import re
import requests
from . import settings
from django.db.models import QuerySet, Case, When
from typing import Union, Any


class PersistanceMixin(object):
    def save(self):
        is_list = isinstance(self.data, list)

        if is_list:
            for item in self.data:
                self.serializer(data=item).save()
        else:
            self.serializer(data=self.data).save()


class BaseHttpDataSet(PersistanceMixin):
    def __init__(self, data, serializer):
        self.data = data
        self.serializer = serializer


class HttpDataSet(BaseHttpDataSet, PersistanceMixin):
    pass


class ProxyResponse(object):
    def __init__(self, data_set, status):
        self.data_set = data_set
        self.status = status


class HttpService(object):
    """
        Http Service class to perform requests to an external API, fetch the results, validate them in
        the database and simplify the translation into a local queryset.
    """
    model = None
    serializer_class = None
    api_url: str = settings.EXTERNAL_API_URL
    authorization_header: str = settings.AUTHORIZATION_HEADER
    authorization_header_name: str = settings.AUTHORIZATION_HEADER_NAME
    primary_key_field_name: str = 'pk'

    def __init__(self):
        self.token = None
        self.headers = dict()
        self.http = requests

    def get_token(self) -> str:
        return self.token

    def set_token(self, token: str):
        self.token = token

    def get_serializer_class(self):
        if not self.model:
            raise Exception('You need to override .get_serializer_class() method or '
                            'provide a default serializer_class.')

        return self.serializer_class

    def set_serializer_class(self, serializer_class):
        self.serializer_class = serializer_class

    def get_model(self):
        if not self.model:
            raise Exception('You need to override .get_model() method or provide a default model.')

        return self.model

    def pluralize(self, text: str) -> str:
        """
            Pluralizes a word depending on the last letter of the word
        """
        return f'{text}ies' if text[-1] == 'y' else f'{text}s'

    def get_model_slug(self) -> str:
        """
            Returns the model slug
        """
        slug = re.sub('(?<=.)([A-Z]+)', '-\\1', self.get_model().__name__).lower()
        return self.pluralize(slug)

    def get_url(self, *url_params) -> str:
        """
            Returns the url based on the url params
        """
        params = (self.api_url, self.get_model_slug(), *url_params)
        return '/'.join([str(param) for param in params if param != ''])

    def get_headers(self) -> dict:
        """
            Returns the headers to perform the request to the server
        :return:
        """
        token = self.get_token()
        if token:
            headers = {
                self.authorization_header_name: f'{self.authorization_header} {token}',
                **self.headers,
            }
            return headers
        return self.headers

    def serialize(self, instance, many=False):
        """
            Returns the serialized data given the model instance
        """
        serializer = self.get_serializer_class()
        serialized = serializer(instance, many=many)
        serialized.is_valid(raise_exception=True)

        return serialized.validated_data

    def deserialize(self, data, many=False) -> Union[dict, list, QuerySet]:
        """
            Returns the deserialized data given the data
        """
        serializer = self.get_serializer_class()
        return serializer(data=data, many=many).validated_data

    def get_response_data(self, response):
        if response.status_code not in range(200, 400):
            raise Exception(f'Error {response.status_code}')

        data, is_list = self.map_response(response.json())

        serializer = self.get_serializer_class()
        results = serializer(data=data, many=is_list)
        results.is_valid(raise_exception=True)

        return data

    def map_response(self, data: Union[dict, list]) -> (Union[dict, list], bool):
        """
            Maps the response to a custom format
        """
        is_list = isinstance(data, list)
        return data, is_list

    def get_queryset(self, data: list, exact_order=True, as_pk_list=False) -> Union[list, QuerySet]:
        """
            Returns the queryset given the received data
        """
        pks = [item[self.primary_key_field_name] for item in data]

        if as_pk_list:
            return pks

        q = {
            f'{self.primary_key_field_name}__in': pks
        }
        results = self.get_model().objects.filter(**q)

        if exact_order:
            order = Case(*[When(pk=pk, then=pos) for pos, pk in enumerate(pks)])
            results = results.order_by(order)

        return results

    def list(self, **params) -> ProxyResponse:
        """
            Retrieves a list of items
        :param params: Query params for the url
        :return: JSON response as a dict
        """
        url = self.get_url()

        return self.get(url, **params)

    def retrieve(self, pk: Any, **params) -> ProxyResponse:
        """
            Retrieves an item given its pk or uid
        :param pk: Unique ID of the item
        :param params: Extra query params
        :return: JSON response as a dict
        """
        url = self.get_url(pk)

        return self.get(url, **params)

    def get(self, url: str, **params) -> ProxyResponse:
        """
            Performs a GET request to a server URL
        :param url: The URL
        :param params: Aditional query params
        :return: JSON response as a dict
        """
        response = self.http.get(url, headers=self.get_headers(), params=params)
        data = self.get_response_data(response)
        http_queryset = HttpDataSet(data=data, serializer=self.get_serializer_class())

        return ProxyResponse(data_set=http_queryset, status=response.status_code)

    def post(self, data: dict, query: dict = None) -> ProxyResponse:
        """
            Performs a POST request to the server
        :param data: Data attached
        :param query: Extra querystring as a dict
        :return: JSON response as a dict
        """
        response = self.http.post(self.get_url(), data=data, headers=self.get_headers(), params=query)
        data = self.get_response_data(response)
        http_queryset = HttpDataSet(data=data, serializer=self.get_serializer_class())

        return ProxyResponse(data_set=http_queryset, status=response.status_code)

    def create(self, data: dict, query: dict = None) -> ProxyResponse:
        return self.post(data=data, query=query)

    def put(self, pk: Any, data: dict, query: dict = None) -> ProxyResponse:
        """
            Performs a PUT request to the server
        :param pk: Primary key of the object to update
        :param data: Updated data
        :param query: Query params
        :return: JSON response as a dict
        """
        response = self.http.put(self.get_url(pk), data=data, headers=self.get_headers(), params=query)
        data = self.get_response_data(response)
        http_queryset = HttpDataSet(data=data, serializer=self.get_serializer_class())

        return ProxyResponse(data_set=http_queryset, status=response.status_code)

    def update(self, pk: Any, data: dict, query: dict = None) -> ProxyResponse:
        return self.put(pk=pk, data=data, query=query)

    def delete(self, pk: Any, query: dict = None) -> ProxyResponse:
        """
            Performs a DELETE request to the server
        :param pk: Primary key of the object to update
        :param query: Query params
        :return: JSON response as a dict
        """
        response = self.http.delete(self.get_url(pk), headers=self.get_headers(), params=query)
        data = self.get_response_data(response)
        http_queryset = HttpDataSet(data=data, serializer=self.get_serializer_class())

        return ProxyResponse(data_set=http_queryset, status=response.status_code)

    def destroy(self, pk: Any, query: dict = None) -> ProxyResponse:
        return self.delete(pk=pk, query=query)



from django.conf import settings

EXTERNAL_API_URL = getattr('SPOOK_API_URL', settings, '')
AUTHORIZATION_HEADER_NAME = getattr('SPOOK_AUTHORIZATION_HEADER_NAME', settings, 'Authorization')
AUTHORIZATION_HEADER = getattr('SPOOK_AUTHORIZATION_HEADER', settings, 'Bearer')



from rest_framework.generics import (
    ListAPIView, RetrieveAPIView, CreateAPIView, UpdateAPIView, DestroyAPIView
)
from rest_framework.response import Response
from .services import HttpService


class BaseHttpServiceView(object):
    service: 'HttpService' = None

    def get_service(self):
        if not self.service:
            raise Exception('You have to specify the service property or override .get_service() function')

        return self.service


class HttpServiceListView(ListAPIView, BaseHttpServiceView):
    def list(self, request, *args, **kwargs):
        service = self.get_service()
        params = request.query_params
        response = service.list(**params)

        return Response(data=response.data_set.data, status=response.status)


class HttpServiceRetrieveView(RetrieveAPIView, BaseHttpServiceView):
    def retrieve(self, request, *args, **kwargs):
        pk = kwargs.get(self.lookup_field)
        service = self.get_service()
        params = request.query_params
        response = service.retrieve(pk, **params)

        return Response(data=response.data_set.data, status=response.status)


class HttpServiceCreateView(CreateAPIView, BaseHttpServiceView):
    def create(self, request, *args, **kwargs):
        service = self.get_service()
        response = service.post(data=request.data, query=request.query_params)

        return Response(data=response.data_set.data, status=response.status)


class HttpServicePutView(UpdateAPIView, BaseHttpServiceView):
    def update(self, request, *args, **kwargs):
        pk = kwargs.get(self.lookup_field)
        service = self.get_service()
        response = service.put(pk=pk, data=request.data, query=request.query_params)

        return Response(data=response.data_set.data, status=response.status)


class HttpServiceDestroyView(DestroyAPIView, BaseHttpServiceView):
    def destroy(self, request, *args, **kwargs):
        pk = kwargs.get(self.lookup_field)
        service = self.get_service()
        response = service.delete(pk=pk, query=request.query_params)

        return Response(data=response.data_set.data, status=response.status)


class HttpServiceRetrieveUpdateView(HttpServiceRetrieveView, HttpServicePutView):
    pass


class HttpServiceRetrieveUpdateDestroyView(HttpServiceRetrieveUpdateView, HttpServiceDestroyView):
    pass


class HttpServiceListCreateView(HttpServiceListView, HttpServiceCreateView):
    pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['spook']

package_data = \
{'': ['*']}

install_requires = \
['django>=1.11.0',
 'djangorestframework>=3.11.0,<4.0.0',
 'requests>=2.24.0,<3.0.0']

setup_kwargs = {
    'name': 'spook',
    'version': '0.1.0rc0',
    'description': 'Django Rest Framework library to interconnect external APIs',
    'long_description': "# Django Spook\n\nLibrary to interconnect multiple external HTTP APIs as Http Services\n\n## Installation\n\n```bash\npip install django-spook\n```\n\n## Usage\n\nDeclare your internal model\n\n```python\nclass MyModel(models.Model):\n    name = models.CharField(max_length=16)\n    age = models.IntegerField(default=0)\n```\n\nDeclare a serializer class for your external service\n\n```python\nfrom rest_framework import serializers\n\nclass MyModelSerializer(serializers.ModelSerializer):\n    class Meta:\n        model = MyModel\n        fields = ('name', 'age', )\n```\n\nDeclare a Http Service class\n\n```python\nfrom spook.services import HttpService\n\nclass MyService(HttpService):\n    model = MyModel\n    serializer_class = MyModelSerializer\n    api_url = 'https://my.external/api'\n```\n\nAnd you can instance MyService class and use the methods\n\n```python\nservice = MyService()\n\nresponse = service.list()\ndata = response.data_set.data\nservice.get_queryset(data=data)\n```",
    'author': 'Pablo Moreno',
    'author_email': 'pablomoreno.inf@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
